_ENV = AIContext

function is_valid_target(ship_or_side, target)
  return target and not (target.dead or target.defenseless) and enemies2(ship_or_side, target)
end

local function is_combat_ship(ship)
  for w in weapons(ship)
    if alive_weapons(w) > 0
      return true
    end
  end
  return false
end

-- don't count ships whose weapons are destroyed, as combat ships
function enemy_combat_ships()
  local f, s, v0 = current_enemy_ships()
  return function(s,v)
    local a = f(s,v)
    while a and not is_combat_ship(a)
      a=f(s,a)
    end
    return a
  end, s, v0
end

function enemies2(a,b)
  if is_object(a)
    a = a.empire
  end
  if is_object(b)
    b = b.empire
  end
  return Battle.enemy(a,b)
end

-- TODO factor in weapon mods if needed
-- average damage per mount
function average_weapon_damage(weapon)
  if weapon.is_fighter
    --while hangars bays are not counted for attack value, handling hangars here
    --is important for deciding whether or not to reserve interceptors
    local ship = weapon.object
    local fighter_weapon = get_fighter_weapon(weapon.parent_type, ship.empire, ship)
    return WEAPONS[fighter_weapon].damage_fun.average
  end

  return weapon.info.damage_fun.average
end

-- average damage per subobject
function average_tracker_damage(tracker)
  --print("TRACKER", tracker.name, tracker.is_missile, tracker.is_fighter)
  if(tracker.is_missile)
    if(tracker.submunitions > 0)
      return WEAPONS[tracker.submunitions_weapon].damage_fun.average * tracker.submunitions
    end
    return WEAPONS[tracker.warhead.parent_type].damage_fun.average
  end
  
  if(tracker.is_fighter)
    local total = 0
    --fighters only ever have 1 weapon equipped at a time right now, but loop through weapons just in case...
    for _,wname in pairs(tracker.weapons) do
      total += WEAPONS[wname.parent_type].damage_fun.average
    end
    return total
  end

  return 0
end

function average_tracker_evasion(speed)
  --based on the hit_chance() calculation in beam_attack_rolls.lua
  local evasion = clamp(2*speed, 0, 100)/100
  local evasion_pd = clamp(2*speed - 50, 0, 100)/100
  return (evasion + evasion_pd)/2
end


--[[
local function desired_launcher_range(w, tracker_survival)
  if w.is_anti_missile
    return ranges.anti_missile
  end

  if w.type=='energy_torpedo'
    return 0.8 * w:launcher_range()
  end

  -- this is based on the idea that: tracker damage ~ (PD survival rate)^max(turns in flight, 1)
  -- therefore: max turns in flight = max(1, log(min acceptable damage)/log(survival rate) )

  -- minimum acceptable damage fraction - kinda: if *this* tracker is harder to hit 
  -- we tolerate less 'damage' as calculated by the *average* PD survival rate
  local min_damage = 1 - average_tracker_evasion(w.launcher_speed)

  local max_turns = 1
  if tracker_survival > 0 and min_damage < 1
    max_turns = max(log(min_damage)/log(tracker_survival), 1)
  end

  -- NOTE that launchers originate from their parent's center and need to reach their target's center
  -- in order to hit. So these engagement ranges need to be a bit shorter than the would be otherwise
  -- in order to compensate for that --> multiply by 0.8
  if w.is_fighter
    return min(0.45*(max_turns), 0.8) * w:launcher_range()
  end

  print(..max_turns, ..min_damage, ..tracker_survival)

  return 0.8 * max_turns * w:launcher_range()
end
--]]

-- desired engagement ranges
local function desired_weapon_range(w, tracker_survival)
  tracker_survival = tracker_survival or 1

  if w.is_anti_missile
    return ranges.anti_missile
  end

  if w.type=='energy_torpedo'
    return 0.9 * w:launcher_range()
  end

  if w.is_launcher
    tracker_survival *= average_tracker_evasion(w.launcher_speed)

    if w.is_fighter
      return w:launcher_range()*(0.4 + 0.5*tracker_survival) -- 0.4 to 0.9
    else
      return w:launcher_range()*(0.8 + 3.2*tracker_survival) -- 0.8 to 4
    end
  end

  --passing 1 to effective_distance() makes it return a range divisor
  if w.is_pd or w.is_beam
    return ranges.short/w:effective_distance(1)
  end
  return ranges.medium/w:effective_distance(1)
end

function desired_attack_dist(ship, tracker_survival)
  if ship.weapons?[1]?.parent_type=='Suicide Attack'
    return 3
  end

  local totalweight = 0
  local total = 0

  local totalweight_pd = 0
  local total_pd = 0

  for w in weapons(ship) do
    local alive = alive_weapons(w)
    if alive > 0
      local range = desired_weapon_range(w, tracker_survival)

      local damage = average_weapon_damage(w)
      local weight = damage * damage

      if(w.is_pd)
        total_pd += weight * range
        totalweight_pd += weight
      else
        total += weight * range
        totalweight += weight
      end
    end
  end

  --weighted average of all non-pd weapons, unless this ship only has PD weapons
  if(totalweight == 0)
    total = total_pd
    totalweight = totalweight_pd
  end

  --return nil if this ship doesn't attack
  return (totalweight > 0 and total / totalweight) or nil
end

-- since trackers need to reach their target's center, we have to use 
-- point-to-point distance to determine their range-to-target accurately
function launcher_dist_to_target(carrier, target)
  cx,cy = object_center(carrier)
  tx,ty = object_center(target)
  return point_dist(cx,cy,tx,ty)
end

----- Point defence calculations

-- multiplier applied to the combat power of things that are vulnerable to point defence
function point_defence_mult(ai_data, obj)
  local tracker_survival_rate
  if enemies2(obj, current_side())
    tracker_survival_rate = 1-ai_data.point_def_effectiveness
  else
    tracker_survival_rate = ai_data.estimated_tracker_survival
  end
  return 1 - 0.9*tracker_survival_rate
end

-- returns a survival rate for our trackers and enemy trackers
function evaluate_point_defences(ai_data)

  local side = current_side()

  -- count the total number of trackers we can launch
  -- and figure out our average tracker health
  -- TODO median tracker health might be a better metric
  local friendly_tracker_health, friendly_tracker_count = 0,0
  local hostile_tracker_health, hostile_tracker_count = 0,0

  for obj in objects() do
    if is_ship(obj)
      for w in weapons(obj) do
        if w.is_launcher
          local n = alive_and_present(w)
          if enemies2(obj, side)
            hostile_tracker_health += n * ai_data.defense_value_each[w] --defence value is "effective health", adjusted for evasion
            hostile_tracker_count += n
          else
            friendly_tracker_health += n * ai_data.defense_value_each[w]
            friendly_tracker_count += n
          end
        end
      end
    elseif is_tracking_object(obj)
      local n in obj
      if enemies2(obj, side)
        hostile_tracker_health += n * ai_data.defense_value_each[obj]
        hostile_tracker_count += n
      else
        friendly_tracker_health += n * ai_data.defense_value_each[obj]
        friendly_tracker_count += n
      end
    end
  end
  friendly_tracker_health /= friendly_tracker_count
  hostile_tracker_health /= hostile_tracker_count

  -- rough estimate of the number of trackers the enemy fleet could kill
  local friendly_point_defence = 0
  local hostile_point_defence = 0
  for obj in objects() do
    if is_ship(obj)
      local ship = obj
      for w in weapons(ship) do
        if not w.cant_attack_trackers
          --factor in PD already fired (and thus unable to autofire this turn), but not too much
          --they do get to fire them next turn, after all.
          local usable = useable_weapons(w)
          local alive = max(alive_weapons(w) - w.missing, 0)
          local n = (usable + 2*alive)/3
          local damage = average_weapon_damage(w)

          if not w.rotates
            damage *= 0.87
          end

          if enemies2(obj, side)
            local hits = min(damage/friendly_tracker_health, 1) * n
            hostile_point_defence += hits
          else
            local hits = min(damage/hostile_tracker_health, 1) * n
            friendly_point_defence += hits
          end
        end
      end
    elseif is_tracking_object(obj)
      --no real way to tell if a tracking object is an 'interceptor', hopefully this is good enough
      if obj.is_fighter and is_tracking_object(obj.dest)
        for _, wname in pairs(obj.weapons) do
          local damage = WEAPONS[wname.parent_type].damage_fun.average

          if enemies2(obj, side)
            hostile_point_defence += min(damage/friendly_tracker_health, 1) * obj.n
          else
            friendly_point_defence += min(damage/hostile_tracker_health, 1) * obj.n
          end
        end
      end
    end
  end

  --estimated tracker survival rate
  local tracker_survial_rate = friendly_tracker_count > 0 and max(friendly_tracker_count - hostile_point_defence, 0)/friendly_tracker_count or 0
  local point_defence_effectiveness = 1 - (hostile_tracker_count > 0 and max(hostile_tracker_count - friendly_point_defence, 0)/hostile_tracker_count or 0)

  return tracker_survial_rate, point_defence_effectiveness
end